import pytest

import gymnasium as gym

from gymcts.gymcts_agent import GymctsAgent
from gymcts.gymcts_deepcopy_wrapper import DeepCopyMCTSGymEnvWrapper


def test_frozenlake_4x4():
    env = gym.make('FrozenLake-v1', desc=None, map_name="4x4", is_slippery=True)
    env.reset()

    # 1. wrap the environment with the naive wrapper or a custom gymcts wrapper
    env = DeepCopyMCTSGymEnvWrapper(env)

    # 2. create the agent
    agent = GymctsAgent(
        env=env,
        clear_mcts_tree_after_step=False,
        render_tree_after_step=True,
        number_of_simulations_per_step=50,
        exclude_unvisited_nodes_from_render=True
    )

    # 3. solve the environment
    actions = agent.solve()

    # 4. render the environment solution in the terminal
    for a in actions:
        obs, rew, term, trun, info = env.step(a)

    episode_return = info["episode"]["r"]

    assert episode_return >= 1.0


@pytest.mark.skip(reason="Sometimes the test fails due to the random nature of the environment.")
def test_cartpole():
    env = gym.make("CartPole-v1")
    env.reset()

    env = DeepCopyMCTSGymEnvWrapper(env)

    agent = GymctsAgent(
        env=env,
        number_of_simulations_per_step=50,
        clear_mcts_tree_after_step=True,
    )

    terminal = False
    step = 0
    while not terminal:
        action, _ = agent.perform_mcts_step()
        obs, rew, term, trun, info = env.step(action)
        terminal = term or trun
        step += 1

    assert step >= 475


def test_mountain_car_continuous():
    env = gym.make("MountainCarContinuous-v0")
    env.reset()

    with pytest.raises(ValueError):
        # continuous action spaces are not supported
        env = DeepCopyMCTSGymEnvWrapper(env)




import random
import copy

import numpy as np
from typing import TypeVar, Any, SupportsFloat, Callable
import gymnasium as gym
from gymnasium.core import WrapperActType, WrapperObsType
from gymnasium.wrappers import RecordEpisodeStatistics

from gymcts.gymcts_env_abc import GymctsABC

from gymcts.logger import log


class DeepCopyMCTSGymEnvWrapper(GymctsABC, gym.Wrapper):
    """
    A wrapper for gym environments that implements the GymctsABC interface.
    It uses deepcopys as state representation.
    Please note that this is not the most efficient way to implement the state representation.
    It is supposed to be used to see if your use-case works well with the MCTS algorithm.
    If it does, you can consider implementing all GymctsABC methods in a more efficient way.
    """

    # helper attributes for the wrapper
    _terminal_flag:bool = False
    _last_reward: SupportsFloat = 0
    _step_tuple: tuple[WrapperObsType, SupportsFloat, bool, bool, dict[str, Any]] = None

    _action_mask_fn: Callable[[gym.Env], np.ndarray] | None = None

    def is_terminal(self) -> bool:
        """
        Returns True if the environment is in a terminal state, False otherwise.

        :return: True if the environment is in a terminal state, False otherwise.
        """
        return self._terminal_flag

    def load_state(self, state: Any) -> None:
        """
        The load_state method is not implemented. The state is loaded by replacing the env with the 'state' (the copy
        provided my 'get_state'). 'self' in a method cannot be replaced with another object (as far as i know).

        :param state: a deepcopy of the environment
        :return: None
        """
        msg = """
        The NaiveSoloMCTSGymEnvWrapper uses deepcopies of the entire env as the state.
        The loading of the state is done by replacing the env with the 'state' (the copy provided my 'get_state').
        'self' in a method cannot be replaced with another object (as far as i know). Therefore the copy is done by
        MCTSaAgent here.
        """
        raise NotImplementedError(msg)

    def __init__(self,
                 env,
                 action_mask_fn: str | Callable[[gym.Env], np.ndarray] | None = None,
                 buffer_length: int = 100,
                 record_video: bool = False,
                 ):
        """
        The constructor of the wrapper. It wraps the environment with RecordEpisodeStatistics and checks if the action
        space is discrete. It also checks if the action_mask_fn is a string or a callable. If it is a string, it tries to
        find the method in the environment. If it is a callable, it assigns it to the _action_mask_fn attribute.

        :param env: the environment to wrap
        :param action_mask_fn:
        :param buffer_length:
        :param record_video:
        """
        # wrap with RecordEpisodeStatistics if it is not already wrapped
        env = RecordEpisodeStatistics(env, buffer_length=buffer_length)

        gym.Wrapper.__init__(self, env)
        # super().__init__(env)

        # assert that the action space is discrete
        if not isinstance(env.action_space, gym.spaces.Discrete):
            raise ValueError("Only discrete action spaces are supported.")

        if action_mask_fn is not None:
            # copy of stable baselines3 contrib implementation
            if isinstance(action_mask_fn, str):
                found_method = getattr(self.env, action_mask_fn)
                if not callable(found_method):
                    raise ValueError(f"Environment attribute {action_mask_fn} is not a method")

                self._action_mask_fn = found_method
            else:
                self._action_mask_fn = action_mask_fn

    def get_state(self) -> Any:
        """
        Returns the current state of the environment as a deepcopy of the environment.
        :return: a deepcopy of the environment
        """
        log.debug("getting state")
        original_state = self
        copied_state = copy.deepcopy(self)

        log.debug(f"original state memory location: {hex(id(original_state))}")
        log.debug(f"copied memory location: {hex(id(copied_state))}")

        return copied_state

    def action_masks(self) -> np.ndarray | None:
        """
        Returns the action masks for the environment. If the action_mask_fn is not set, it returns None.
        :return: the action masks for the environment
        """
        return self._action_mask_fn(self.env) if self._action_mask_fn is not None else None

    def get_valid_actions(self) -> list[int]:
        """
        Returns a list of valid actions for the current state of the environment.
        This used to obtain potential actions/subsequent sates for the MCTS tree.

        :return: the list of valid actions
        """
        if self._action_mask_fn is None:
            action_space: gym.spaces.Discrete = self.env.action_space  # Type hinting
            return list(range(action_space.n))
        else:
            return [i for i, mask in enumerate(self.action_masks()) if mask]

    def step(
            self, action: WrapperActType
    ) -> tuple[WrapperObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        """
        Performs a step in the environment.
        This method is used to update the wrapper with the new state and the new action, to realize the terminal state
        functionality.

        :param action: action to perform in the environment
        :return: the step tuple of the environment (obs, reward, terminated, truncated, info)
        """
        step_tuple = self.env.step(action)

        obs, reward, terminated, truncated, info = step_tuple
        self._terminal_flag = terminated or truncated
        self._step_tuple = step_tuple

        return step_tuple


    def rollout(self) -> float:
        """
        Performs a rollout from the current state of the environment and returns the return (sum of rewards) of the
        rollout.

        :return: the return of the rollout
        """
        log.debug("performing rollout")
        # random rollout
        # perform random valid action util terminal
        is_terminal_state = self.is_terminal()

        if is_terminal_state:
            _, _, _, _, info = self._step_tuple
            episode_return = info["episode"]["r"]
            return episode_return

        while not is_terminal_state:
            action = random.choice(self.get_valid_actions())
            # print(f"Valid actions: {self.get_valid_actions()}, selected action: {action}")
            _obs, _reward, is_terminal_state, _truncated, info = self.step(action)


        episode_return = info["episode"]["r"]
        log.debug(f"Rollout return: {episode_return}")
        return episode_return



import copy
import random
import gymnasium as gym

from typing import TypeVar, Any, SupportsFloat, Callable, Literal

from gymcts.gymcts_env_abc import GymctsABC
from gymcts.gymcts_deepcopy_wrapper import DeepCopyMCTSGymEnvWrapper
from gymcts.gymcts_node import GymctsNode
from gymcts.gymcts_tree_plotter import _generate_mcts_tree

from gymcts.logger import log






class GymctsAgent:
    render_tree_after_step: bool = False
    render_tree_max_depth: int = 2
    exclude_unvisited_nodes_from_render: bool = False
    number_of_simulations_per_step: int = 25

    env: GymctsABC
    search_root_node: GymctsNode  # NOTE: this is not the same as the root of the tree!
    clear_mcts_tree_after_step: bool


    # (num_simulations: int, step_idx: int) -> int
    @staticmethod
    def calc_number_of_simulations_per_step(num_simulations: int, step_idx: int) -> int:
        """
        A function that returns a constant number of simulations per step.

        :param num_simulations: The number of simulations to return.
        :param step_idx: The current step index (not used in this function).
        :return: A callable that takes an environment as input and returns the constant number of simulations.
        """
        return num_simulations

    def __init__(self,
                 env: GymctsABC,
                 clear_mcts_tree_after_step: bool = True,
                 render_tree_after_step: bool = False,
                 render_tree_max_depth: int = 2,
                 number_of_simulations_per_step: int = 25,
                 exclude_unvisited_nodes_from_render: bool = False,
                 calc_number_of_simulations_per_step: Callable[[int,int], int] = None,
                 score_variate: Literal["UCT_v0", "UCT_v1", "UCT_v2",] = "UCT_v0",
                 best_action_weight=None,
                 keep_whole_tree_till_initial_root: bool = False,
                 ):
        # check if action space of env is discrete
        if not isinstance(env.action_space, gym.spaces.Discrete):
            raise ValueError("Action space must be discrete.")
        if calc_number_of_simulations_per_step is not None:
            # check if the provided function is callable
            if not callable(calc_number_of_simulations_per_step):
                raise ValueError("calc_number_of_simulations_per_step must be a callable accepting two arguments: num_simulations and step_idx.")
            # assign the provided function to the attribute
            # it needs to be staticmethod to be used as a class attribute
            print("Using provided calc_number_of_simulations_per_step function.")
            self.calc_number_of_simulations_per_step = staticmethod(calc_number_of_simulations_per_step)
        if score_variate not in ["UCT_v0", "UCT_v1", "UCT_v2"]:
            raise ValueError("score_variate must be one of ['UCT_v0', 'UCT_v1', 'UCT_v2'].")
        GymctsNode.score_variate = score_variate

        if best_action_weight is not None:
            if best_action_weight < 0 or best_action_weight > 1:
                raise ValueError("best_action_weight must be in range [0, 1].")
            GymctsNode.best_action_weight = best_action_weight


        self.render_tree_after_step = render_tree_after_step
        self.exclude_unvisited_nodes_from_render = exclude_unvisited_nodes_from_render
        self.render_tree_max_depth = render_tree_max_depth

        self.number_of_simulations_per_step = number_of_simulations_per_step

        self.env = env
        self.clear_mcts_tree_after_step = clear_mcts_tree_after_step
        self.keep_whole_tree_till_initial_root = keep_whole_tree_till_initial_root

        self.search_root_node = GymctsNode(
            action=None,
            parent=None,
            env_reference=env,
        )

    def navigate_to_leaf(self, from_node: GymctsNode) -> GymctsNode:
        log.debug(f"Navigate to leaf. from_node: {from_node}")
        if from_node.terminal:
            log.debug("Node is terminal. Returning from_node")
            return from_node
        if from_node.is_leaf():
            log.debug("Node is leaf. Returning from_node")
            return from_node

        temp_node = from_node
        # NAVIGATION STRATEGY
        # select child with highest UCB score
        while not temp_node.is_leaf():
            children = list(temp_node.children.values())
            max_ucb_score = max(child.tree_policy_score() for child in children)
            best_children = [child for child in children if child.tree_policy_score() == max_ucb_score]
            temp_node = random.choice(best_children)
        log.debug(f"Selected leaf node: {temp_node}")
        return temp_node

    def expand_node(self, node: GymctsNode) -> None:
        log.debug(f"expanding node: {node}")
        # EXPANSION STRATEGY
        # expand all children

        child_dict = {}
        for action in node.valid_actions:
            # reconstruct state
            # load state of leaf node
            self._load_state(node)

            obs, reward, terminal, truncated, _ = self.env.step(action)
            child_dict[action] = GymctsNode(
                action=action,
                parent=node,
                env_reference=self.env,
            )
        node.children = child_dict

    def solve(self, num_simulations_per_step: int = None, render_tree_after_step: bool = None) -> list[int]:

        if num_simulations_per_step is None:
            num_simulations_per_step = self.number_of_simulations_per_step
        if render_tree_after_step is None:
            render_tree_after_step = self.render_tree_after_step

        log.debug(f"Solving from root node: {self.search_root_node}")

        current_node = self.search_root_node

        action_list = []

        idx = 0
        while not current_node.terminal:
            num_sims = self.calc_number_of_simulations_per_step(num_simulations_per_step, idx)

            log.info(f"Performing MCTS step {idx} with {num_sims} simulations.")

            next_action, current_node = self.perform_mcts_step(num_simulations=num_sims,
                                                               render_tree_after_step=render_tree_after_step)
            log.info(f"selected action {next_action} after {num_sims} simulations.")
            action_list.append(next_action)
            log.info(f"current action list: {action_list}")

            idx += 1

        log.info(f"Final action list: {action_list}")
        # restore state of current node
        return action_list

    def _load_state(self, node: GymctsNode) -> None:
        if isinstance(self.env, DeepCopyMCTSGymEnvWrapper):
            self.env = copy.deepcopy(node.state)
        else:
            self.env.load_state(node.state)

    def perform_mcts_step(self, search_start_node: GymctsNode = None, num_simulations: int = None,
                          render_tree_after_step: bool = None) -> tuple[int, GymctsNode]:

        if render_tree_after_step is None:
            render_tree_after_step = self.render_tree_after_step

        if render_tree_after_step is None:
            render_tree_after_step = self.render_tree_after_step

        if num_simulations is None:
            num_simulations = self.number_of_simulations_per_step

        if search_start_node is None:
            search_start_node = self.search_root_node

        action = self.vanilla_mcts_search(
            search_start_node=search_start_node,
            num_simulations=num_simulations,
        )
        next_node = search_start_node.children[action]

        if self.clear_mcts_tree_after_step:
            # to clear memory we need to remove all nodes except the current node
            # this is done by setting the root node to the current node
            # and setting the parent of the current node to None
            # we also need to reset the children of the current node
            # this is done by calling the reset method
            next_node.reset()
        elif not self.keep_whole_tree_till_initial_root:
            next_node.remove_parent()

        self.search_root_node = next_node

        return action, next_node

    def vanilla_mcts_search(self, search_start_node: GymctsNode = None, num_simulations=10) -> int:
        log.debug(f"performing one MCTS search step with {num_simulations} simulations")
        if search_start_node is None:
            search_start_node = self.search_root_node

        for i in range(num_simulations):
            log.debug(f"simulation {i}")
            # navigate to leaf
            leaf_node = self.navigate_to_leaf(from_node=search_start_node)

            if leaf_node.visit_count > 0 and not leaf_node.terminal:
                # expand leaf
                self.expand_node(leaf_node)
                leaf_node = leaf_node.get_random_child()

            # load state of leaf node
            self._load_state(leaf_node)

            # rollout
            episode_return = self.env.rollout()
            # self.env.render()

            self.backpropagation(node=leaf_node, episode_return=episode_return)

        if self.render_tree_after_step:
            self.show_mcts_tree()

        return search_start_node.get_best_action()

    def show_mcts_tree(self, start_node: GymctsNode = None, tree_max_depth: int = None) -> None:

        if start_node is None:
            start_node = self.search_root_node

        if tree_max_depth is None:
            tree_max_depth = self.render_tree_max_depth

        print(start_node.__str__(colored=True, action_space_n=self.env.action_space.n))
        for line in _generate_mcts_tree(
                start_node=start_node,
                depth=tree_max_depth,
                action_space_n=self.env.action_space.n,
        ):
            print(line)

    def show_mcts_tree_from_root(self, tree_max_depth: int = None) -> None:
        self.show_mcts_tree(start_node=self.search_root_node.get_root(), tree_max_depth=tree_max_depth)

    def backpropagation(self, node: GymctsNode, episode_return: float) -> None:
        log.debug(f"performing backpropagation from leaf node: {node}")
        while not node.is_root():
            # node.mean_value = ((node.mean_value * node.visit_count) + episode_return) / (node.visit_count + 1)
            node.mean_value = node.mean_value + (episode_return - node.mean_value) / (node.visit_count + 1)
            node.visit_count += 1
            node.max_value = max(node.max_value, episode_return)
            node.min_value = min(node.min_value, episode_return)
            node = node.parent
        # also update root node
        # node.mean_value = ((node.mean_value * node.visit_count) + episode_return) / (node.visit_count + 1)
        node.mean_value = node.mean_value + (episode_return - node.mean_value) / (node.visit_count + 1)
        node.visit_count += 1
        node.max_value = max(node.max_value, episode_return)
        node.min_value = min(node.min_value, episode_return)





import copy
import gymnasium as gym

from typing import TypeVar, Any, SupportsFloat, Callable

from ray.types import ObjectRef

from gymcts.gymcts_agent import GymctsAgent
from gymcts.gymcts_env_abc import GymctsABC
from gymcts.gymcts_deepcopy_wrapper import DeepCopyMCTSGymEnvWrapper
from gymcts.gymcts_node import GymctsNode
from gymcts.gymcts_tree_plotter import _generate_mcts_tree

from gymcts.logger import log

import ray
import copy

TSoloMCTSNode = TypeVar("TSoloMCTSNode", bound="SoloMCTSNode")


@ray.remote
def mcts_lookahead(
        gymcts_start_node: GymctsNode,
        env: GymctsABC,
        num_simulations: int) -> GymctsNode:
    agent = GymctsAgent(
        env=env,
        clear_mcts_tree_after_step=False,
        number_of_simulations_per_step=num_simulations,
    )
    agent.search_root_node = gymcts_start_node

    agent.vanilla_mcts_search(
        search_start_node=gymcts_start_node,
        num_simulations=num_simulations,
    )
    return agent.search_root_node


def merge_nodes(gymcts_node1, gymcts_node2, perform_state_equality_check=False):
    log.debug(f"merging {gymcts_node1} and {gymcts_node2}")
    # maybe add some state equality check here
    if perform_state_equality_check:
        if gymcts_node1.state != gymcts_node2.state:
            raise ValueError("States are different")

    if gymcts_node1 is None:
        log.debug(f"first node is None, returning second node ({gymcts_node2})")
        return gymcts_node2
    if gymcts_node2 is None:
        log.debug(f"second node is None, returning first node ({gymcts_node1})")
        return gymcts_node1
    if gymcts_node1 is None and gymcts_node2 is None:
        log.error("Both nodes are None")
        raise ValueError("Both nodes are None")

    if gymcts_node1.is_leaf() and not gymcts_node2.is_leaf():
        log.debug(f"first node is leaf, second node is not leaf")
        gymcts_node2.parent = gymcts_node1.parent
        log.debug(f"returning first node: {gymcts_node2}")
        return gymcts_node2

    if gymcts_node2.is_leaf() and not gymcts_node1.is_leaf():
        log.debug(f"second node is leaf, first node is not leaf")
        log.debug(f"returning first node: {gymcts_node1}")
        return gymcts_node1

    if gymcts_node1.is_leaf() and gymcts_node2.is_leaf():
        log.debug(f"both nodes are leafs, returning first node")
        log.debug(f"returning first node: {gymcts_node1}")
        return gymcts_node1

    # check if gymcts_node1 and gymcts_node2 have the same children
    if gymcts_node1.children.keys() != gymcts_node2.children.keys():
        log.error("Nodes have different children")
        raise ValueError("Nodes have different children")

    for (action1, child1), (action2, child2) in zip(gymcts_node1.children.items(), gymcts_node2.children.items()):
        if action1 != action2:
            log.error("Actions are different")
            raise ValueError("Actions are different")
        log.debug(f"merging children with action {action1} for node {gymcts_node1}")
        gymcts_node1.children[action1] = merge_nodes(
            child1,
            child2,
            perform_state_equality_check=perform_state_equality_check
        )

    visit_count = gymcts_node1.visit_count + gymcts_node2.visit_count
    mean_value = (
                         gymcts_node1.mean_value * gymcts_node1.visit_count + gymcts_node2.mean_value * gymcts_node2.visit_count) / visit_count
    max_value = max(gymcts_node1.max_value, gymcts_node2.max_value)
    min_value = min(gymcts_node1.min_value, gymcts_node2.min_value)

    gymcts_node1.visit_count = visit_count
    gymcts_node1.mean_value = mean_value
    gymcts_node1.max_value = max_value
    gymcts_node1.min_value = min_value
    log.debug(f"merged node: {gymcts_node1}")
    log.debug(f"returning node: {gymcts_node1}")
    return gymcts_node1


class DistributedGymctsAgent:
    render_tree_after_step: bool = False
    render_tree_max_depth: int = 2
    exclude_unvisited_nodes_from_render: bool = False
    number_of_simulations_per_step: int = 25
    num_parallel: int = 4

    env: GymctsABC
    search_root_node: GymctsNode  # NOTE: this is not the same as the root of the tree!
    clear_mcts_tree_after_step: bool

    def __init__(self,
                 env: GymctsABC,
                 render_tree_after_step: bool = False,
                 render_tree_max_depth: int = 2,
                 num_parallel: int = 4,
                 clear_mcts_tree_after_step: bool = False,
                 number_of_simulations_per_step: int = 25,
                 exclude_unvisited_nodes_from_render: bool = False
                 ):
        # check if action space of env is discrete
        if not isinstance(env.action_space, gym.spaces.Discrete):
            raise ValueError("Action space must be discrete.")

        self.num_parallel = num_parallel

        self.render_tree_after_step = render_tree_after_step
        self.exclude_unvisited_nodes_from_render = exclude_unvisited_nodes_from_render
        self.render_tree_max_depth = render_tree_max_depth

        self.number_of_simulations_per_step = number_of_simulations_per_step

        self.env = env
        self.clear_mcts_tree_after_step = clear_mcts_tree_after_step

        self.search_root_node = GymctsNode(
            action=None,
            parent=None,
            env_reference=env,
        )

    def solve(self, num_simulations_per_step: int = None, render_tree_after_step: bool = None) -> list[int]:

        if num_simulations_per_step is None:
            num_simulations_per_step = self.number_of_simulations_per_step
        if render_tree_after_step is None:
            render_tree_after_step = self.render_tree_after_step

        log.debug(f"Solving from root node: {self.search_root_node}")

        current_node = self.search_root_node

        action_list = []

        while not current_node.terminal:
            next_action, current_node = self.perform_mcts_step(num_simulations=num_simulations_per_step,
                                                               render_tree_after_step=render_tree_after_step)
            log.info(
                f"selected action {next_action} after {self.num_parallel} x {num_simulations_per_step} simulations.")
            action_list.append(next_action)
            log.info(f"current action list: {action_list}")


        log.info(f"Final action list: {action_list}")
        # restore state of current node
        return action_list

    def perform_mcts_step(self, search_start_node: GymctsNode = None, num_simulations: int = None,
                          render_tree_after_step: bool = None, num_parallel: int = None) -> tuple[int, GymctsNode]:

        if render_tree_after_step is None:
            render_tree_after_step = self.render_tree_after_step

        if render_tree_after_step is None:
            render_tree_after_step = self.render_tree_after_step

        if num_simulations is None:
            num_simulations = self.number_of_simulations_per_step

        if search_start_node is None:
            search_start_node = self.search_root_node

        if num_parallel is None:
            num_parallel = self.num_parallel

        # action = self.vanilla_mcts_search(
        #    search_start_node=search_start_node,
        #   num_simulations=num_simulations,
        # )
        # next_node = search_start_node.children[action]

        mcts_interation_futures = [
            mcts_lookahead.remote(
                copy.deepcopy(search_start_node),
                copy.deepcopy(self.env),
                num_simulations=num_simulations
            )
            for _ in range(num_parallel)
        ]

        while mcts_interation_futures:
            ready_gymcts_nodes, mcts_interation_futures = ray.wait(mcts_interation_futures)
            for ready_node_ref in ready_gymcts_nodes:
                ready_node = ray.get(ready_node_ref)

                # merge the tree
                if not self.clear_mcts_tree_after_step:
                    self.backpropagation(search_start_node, ready_node.mean_value, ready_node.visit_count)
                search_start_node = merge_nodes(search_start_node, ready_node)

        action = search_start_node.get_best_action()
        next_node = search_start_node.children[action]

        if self.render_tree_after_step:
            self.show_mcts_tree(
                start_node=search_start_node,
                tree_max_depth=self.render_tree_max_depth
            )

        if self.clear_mcts_tree_after_step:
            # to clear memory we need to remove all nodes except the current node
            # this is done by setting the root node to the current node
            # and setting the parent of the current node to None
            # we also need to reset the children of the current node
            # this is done by calling the reset method
            next_node.reset()

        self.search_root_node = next_node

        return action, next_node

    def backpropagation(self, node: GymctsNode, average_episode_return: float, num_episodes: int) -> None:
        log.debug(f"performing backpropagation from leaf node: {node}")
        while not node.is_root():
            node.mean_value = (node.mean_value * node.visit_count + average_episode_return * num_episodes) / (
                        node.visit_count + num_episodes)
            node.visit_count += num_episodes
            node.max_value = max(node.max_value, average_episode_return)
            node.min_value = min(node.min_value, average_episode_return)
            node = node.parent
        # also update root node
        node.mean_value = (node.mean_value * node.visit_count + average_episode_return * num_episodes) / (
                    node.visit_count + num_episodes)
        node.visit_count += num_episodes
        node.max_value = max(node.max_value, average_episode_return)
        node.min_value = min(node.min_value, average_episode_return)

    def show_mcts_tree(self, start_node: GymctsNode = None, tree_max_depth: int = None) -> None:

        if start_node is None:
            start_node = self.search_root_node

        if tree_max_depth is None:
            tree_max_depth = self.render_tree_max_depth

        print(start_node.__str__(colored=True, action_space_n=self.env.action_space.n))
        for line in _generate_mcts_tree(
                start_node=start_node,
                depth=tree_max_depth,
                action_space_n=self.env.action_space.n
        ):
            print(line)

    def show_mcts_tree_from_root(self, tree_max_depth: int = None) -> None:
        self.show_mcts_tree(start_node=self.search_root_node.get_root(), tree_max_depth=tree_max_depth)


if __name__ == '__main__':
    ray.init()

    log.setLevel(20)  # 10=DEBUG, 20=INFO, 30=WARNING, 40=ERROR, 50=CR
    env = gym.make('FrozenLake-v1', desc=None, map_name="4x4", is_slippery=False)
    env.reset()

    # 1. wrap the environment with the naive wrapper or a custom gymcts wrapper
    # env1 = ActionHistoryMCTSGymEnvWrapper(env1)
    env = DeepCopyMCTSGymEnvWrapper(env)

    # 2. create the agent
    agent1 = DistributedGymctsAgent(
        env=env,
        render_tree_after_step=True,
        number_of_simulations_per_step=10,
        exclude_unvisited_nodes_from_render=True,
        num_parallel=1,
    )
    import time

    start_time = time.perf_counter()
    actions = agent1.solve()
    end_time = time.perf_counter()

    agent1.show_mcts_tree_from_root()

    print(f"solution time pro action: {end_time - start_time}/{len(actions)}")



import random

import numpy as np
from typing import Any, SupportsFloat, Callable
import gymnasium as gym
from gymnasium.core import WrapperActType, WrapperObsType
from gymnasium.wrappers import RecordEpisodeStatistics

from gymcts.gymcts_env_abc import GymctsABC

from gymcts.logger import log


class ActionHistoryMCTSGymEnvWrapper(GymctsABC, gym.Wrapper):
    """
    A wrapper for gym environments that implements the GymctsABC interface.
    It uses the action history as state representation.
    Please note that this is not the most efficient way to implement the state representation.
    It is supposed to be used to see if your use-case works well with the MCTS algorithm.
    If it does, you can consider implementing all GymctsABC methods in a more efficient way.
    The action history is a list of actions taken in the environment.
    The state is represented as a list of actions taken in the environment.
    The state is used to restore the environment using the load_state method.

    It is supposed to be used to see if your use-case works well with the MCTS algorithm.
    If it does, you can consider implementing all GymctsABC methods in a more efficient way.
    """

    # helper attributes for the wrapper
    _terminal_flag: bool = False
    _last_reward: SupportsFloat = 0
    _step_tuple: tuple[WrapperObsType, SupportsFloat, bool, bool, dict[str, Any]] = None

    _action_mask_fn: Callable[[gym.Env], np.ndarray] | None = None

    def __init__(
            self,
            env,
            action_mask_fn: str | Callable[[gym.Env], np.ndarray] | None = None,
            buffer_length: int = 100,
    ):
        """
        A wrapper for gym environments that implements the GymctsABC interface.
        It uses the action history as state representation.
        Please note that this is not the most efficient way to implement the state representation.
        It is supposed to be used to see if your use-case works well with the MCTS algorithm.
        If it does, you can consider implementing all GymctsABC methods in a more efficient way.

        :param env: the environment to wrap
        :param action_mask_fn: a function that takes the environment as input and returns a mask of valid actions
        :param buffer_length: the length of the buffer for recording episodes for determining their rollout returns
        """
        # wrap with RecordEpisodeStatistics if it is not already wrapped
        env = RecordEpisodeStatistics(env, buffer_length=buffer_length)

        gym.Wrapper.__init__(self, env)

        self._wrapper_action_history = []

        # assert that the action space is discrete
        if not isinstance(env.action_space, gym.spaces.Discrete):
            raise ValueError("Only discrete action spaces are supported.")

        if action_mask_fn is not None:
            # copy of stable baselines3 contrib implementation
            if isinstance(action_mask_fn, str):
                found_method = getattr(self.env, action_mask_fn)
                if not callable(found_method):
                    raise ValueError(f"Environment attribute {action_mask_fn} is not a method")

                self._action_mask_fn = found_method
            else:
                self._action_mask_fn = action_mask_fn

    def load_state(self, state: list[int]) -> None:
        """
        Loads the state of the environment. The state is a list of actions taken in the environment.

        The environment is reset and all actions in the state are performed in order to restore the environment to the
        same state.

        This works only for deterministic environments!

        :param state: the state to load
        :return: None
        """
        self.env.reset()
        self._wrapper_action_history = []

        for action in state:
            self.env.step(action)
            self._wrapper_action_history.append(action)

    def is_terminal(self) -> bool:
        """
        Returns True if the environment is in a terminal state, False otherwise.

        :return:
        """
        if not len(self.get_valid_actions()):
            return True
        else:
            return self._terminal_flag

    def action_masks(self) -> np.ndarray | None:
        """
        Returns the action masks for the environment. If the action_mask_fn is not set, it returns None.

        :return:
        """
        return self._action_mask_fn(self.env) if self._action_mask_fn is not None else None

    def get_valid_actions(self) -> list[int]:
        """
        Returns a list of valid actions for the current state of the environment.

        :return: a list of valid actions
        """
        if self._action_mask_fn is None:
            action_space: gym.spaces.Discrete = self.env.action_space  # Type hinting
            return list(range(action_space.n))
        else:
            return [i for i, mask in enumerate(self.action_masks()) if mask]

    def rollout(self) -> float:
        """
        Performs a random rollout from the current state of the environment and returns the return (sum of rewards)
        of the rollout.

        :return: the return of the rollout
        """
        log.debug("performing rollout")
        # random rollout
        # perform random valid action util terminal
        is_terminal_state = self.is_terminal()

        if is_terminal_state:
            _, _, _, _, info = self._step_tuple
            episode_return = info["episode"]["r"]
            return episode_return

        while not is_terminal_state:
            action = random.choice(self.get_valid_actions())
            # print(f"Valid actions: {self.get_valid_actions()}, selected action: {action}")
            _obs, _reward, is_terminal_state, _truncated, info = self.step(action)

        episode_return = info["episode"]["r"]
        log.debug(f"Rollout return: {episode_return}")
        return episode_return

    def get_state(self) -> list[int]:
        """
        Returns the current state of the environment. The state is a list of actions taken in the environment,
        namely all action that have been taken in the environment so far (since the last reset).

        :return: a list of actions taken in the environment
        """

        return self._wrapper_action_history.copy()

    def step(
            self, action: WrapperActType
    ) -> tuple[WrapperObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        """
        Performs a step in the environment. It adds the action to the action history and updates the terminal flag.

        :param action: action to perform in the environment
        :return: the step tuple of the environment (obs, reward, terminated, truncated, info)
        """
        step_tuple = self.env.step(action)
        self._wrapper_action_history.append(action)
        obs, reward, terminated, truncated, info = step_tuple

        self._terminal_flag = terminated or truncated
        self._step_tuple = step_tuple

        return step_tuple



from typing import TypeVar, Any, SupportsFloat, Callable
from abc import ABC, abstractmethod
import gymnasium as gym
import numpy as np


class GymctsABC(ABC, gym.Env):

    @abstractmethod
    def get_state(self) -> Any:
        """
        Returns the current state of the environment. The state can be any datatype in principle, that allows to restore
        the environment to the same state. The state is used to restore the environment unsing the load_state method.

        It's recommended to use a numpy array if possible, as it is easy to serialize and deserialize.

        :return: the current state of the environment
        """
        pass

    @abstractmethod
    def load_state(self, state: Any) -> None:
        """
        Loads the state of the environment. The state can be any datatype in principle, that allows to restore the
        environment to the same state. The state is used to restore the environment unsing the load_state method.

        :param state: the state to load
        :return: None
        """
        pass

    @abstractmethod
    def is_terminal(self) -> bool:
        """
        Returns True if the environment is in a terminal state, False otherwise.
        :return:
        """
        pass

    @abstractmethod
    def get_valid_actions(self) -> list[int]:
        """
        Returns a list of valid actions for the current state of the environment.
        This used to obtain potential actions/subsequent sates for the MCTS tree.
        :return: the list of valid actions
        """
        pass

    @abstractmethod
    def action_masks(self) -> np.ndarray | None:
        """
        Returns a numpy array of action masks for the environment. The array should have the same length as the number
        of actions in the action space. If an action is valid, the corresponding mask value should be 1, otherwise 0.
        If no action mask is available, it should return None.

        :return: a numpy array of action masks or None
        """
        pass

    @abstractmethod
    def rollout(self) -> float:
        """
        Performs a rollout from the current state of the environment and returns the return (sum of rewards) of the rollout.

        Please make sure the return value is in the interval [-1, 1].
        Otherwise, the MCTS algorithm will not work as expected (due to a male-fitted exploration coefficient;
        exploration and exploitation are not well-balanced then).

        :return: the return of the rollout
        """
        pass



from jsp_instance_utils.instances import ft06, ft06_makespan

from gymcts.gymcts_agent import GymctsAgent


def test_solution_quality_small_instance(graph_matrix_env_naive_wrapper_two_job_jsp_instance):
    env = graph_matrix_env_naive_wrapper_two_job_jsp_instance

    agent = GymctsAgent(env=env)

    actions = agent.solve(num_simulations_per_step=100)  # enough to do a full search
    for a in actions:
        obs, rew, term, trun, info = env.step(a)

    assert env.unwrapped.get_makespan() == 40


def test_solution_quality_ft06_with_():
    import gymnasium as gym
    import numpy as np

    from graph_matrix_jsp_env.disjunctive_jsp_env import DisjunctiveGraphJspEnv
    from gymcts.gymcts_deepcopy_wrapper import DeepCopyMCTSGymEnvWrapper

    env = DisjunctiveGraphJspEnv(
        jsp_instance=ft06,
        reward_function="makespan",
        c_lb=-ft06_makespan,
    )
    env.reset()

    def mask_fn(env: gym.Env) -> np.ndarray:
        return env.unwrapped.valid_action_mask()

    env = DeepCopyMCTSGymEnvWrapper(
        env,
        action_mask_fn=mask_fn
    )

    agent = GymctsAgent(
        env=env,
        number_of_simulations_per_step=200,
        clear_mcts_tree_after_step=False,
    )

    actions = agent.solve()

    for a in actions:
        obs, rew, term, trun, info = env.step(a)

    assert term
    assert env.unwrapped.get_makespan() <= ft06_makespan * 1.5



from typing import Any, Generator

from gymcts.gymcts_node import GymctsNode

from gymcts.logger import log


def _generate_mcts_tree(
        start_node: GymctsNode = None,
        prefix: str = None,
        depth: int = None,
        exclude_unvisited_nodes_from_render: bool = True,
        action_space_n: int = None
) -> Generator[str, Any | None, None]:
    """
    Generates a tree representation of the MCTS tree starting from the given node.

    This is a recursive function that generates a tree representation of the MCTS tree starting from the given node. The

    :param start_node: the node to start from
    :param prefix: used to format the tree
    :param depth: used to limit the depth of the tree
    :param exclude_unvisited_nodes_from_render: used to exclude unvisited nodes from the render
    :param action_space_n: the number of actions in the action space
    :return: a list of strings representing the tree
    """
    if prefix is None:
        prefix = ""
    import gymcts.colorful_console_utils as ccu

    if start_node is None:
        raise ValueError("start_node must not be None")

    if action_space_n is None:
        log.warning("action_space_n is None, defaulting to 100")
        action_space_n = 100

    # prefix components:
    space = '    '
    branch = '│   '
    # pointers:
    tee = '├── '
    last = '└── '

    contents = start_node.children.values() if start_node.children is not None else []
    if exclude_unvisited_nodes_from_render:
        contents = [node for node in contents if node.visit_count > 0]
    # contents each get pointers that are ├── with a final └── :
    # pointers = [tee] * (len(contents) - 1) + [last]
    pointers = [tee for _ in range(len(contents) - 1)] + [last]

    for pointer, current_node in zip(pointers, contents):
        n_item = current_node.parent.action if current_node.parent is not None else 0
        n_classes = action_space_n

        pointer = ccu.wrap_evenly_spaced_color(
            s=pointer,
            n_of_item=n_item,
            n_classes=n_classes,
        )

        yield prefix + pointer + f"{current_node.__str__(colored=True, action_space_n=n_classes)}"
        if current_node.children and len(current_node.children):  # extend the prefix and recurse:
            # extension = branch if pointer == tee else space
            extension = branch if tee in pointer else space
            # i.e. space because last, └── , above so no more |
            extension = ccu.wrap_evenly_spaced_color(
                s=extension,
                n_of_item=n_item,
                n_classes=n_classes,
            )
            if depth is not None and depth <= 0:
                continue
            yield from _generate_mcts_tree(
                current_node,
                prefix=prefix + extension,
                action_space_n=action_space_n,
                depth=depth - 1 if depth is not None else None
            )


def show_mcts_tree(
        start_node: GymctsNode = None,
        tree_max_depth: int = None,
        action_space_n: int = None
) -> None:
    """
    Renders the MCTS tree starting from the given node.

    :param start_node: the node to start from
    :param tree_max_depth: the maximum depth of the tree to render
    :param action_space_n: the number of actions in the action space
    """
    print(start_node.__str__(colored=True, action_space_n=action_space_n))
    for line in _generate_mcts_tree(start_node=start_node, depth=tree_max_depth):
        print(line)



import uuid
import random
import math

from typing import TypeVar, Any, SupportsFloat, Callable, Generator, Literal

from gymcts.gymcts_env_abc import GymctsABC

from gymcts.logger import log

TGymctsNode = TypeVar("TGymctsNode", bound="GymctsNode")


class GymctsNode:
    # static properties
    best_action_weight: float = 0.05 # weight for the best action
    ubc_c = 0.707 # exploration coefficient

    """
    UCT (Upper Confidence Bound applied to Trees) exploration terms:

    UCT 0:
        c * √( 2 * ln(N(s)) / N(s,a) )

    UCT 1:
        c * √( ln(N(s)) / (1 + N(s,a)) )

    UCT 2:
        c * ( √(N(s)) / (1 + N(s,a)) )

    Where:
        N(s)     = number of times state s has been visited
        N(s,a)   = number of times action a was taken from state s
        c        = exploration constant
    """
    score_variate: Literal["UCT_v0", "UCT_v1", "UCT_v2",] = "UCT_v0"



    # attributes
    #
    # Note these attributes are not static. Their defined here to give developers a hint what fields are available
    # in the class. They are not static because they are not shared between instances of the class in scope of
    # this library.
    visit_count: int = 0 # number of times the node has been visited
    mean_value: float = 0 # mean value of the node
    max_value: float = -float("inf") # maximum value of the node
    min_value: float = +float("inf") # minimum value of the node
    terminal: bool = False # whether the node is terminal or not
    state: Any = None # state of the node

    def __str__(self, colored=False, action_space_n=None) -> str:
        """
        Returns a string representation of the node. The string representation is used for visualisation purposes.
        It is used for example in the mcts tree visualisation functionality.

        :param colored: true if the string representation should be colored, false otherwise. (ture is used by the mcts tree visualisation)
        :param action_space_n: the number of actions in the action space. This is used for coloring the action in the string representation.
        :return: a potentially colored string representation of the node.
        """
        if not colored:

            if not self.is_root():
                return f"(a={self.action}, N={self.visit_count}, Q_v={self.mean_value:.2f}, best={self.max_value:.2f}, {GymctsNode.score_variate}={self.tree_policy_score():.2f})"
            else:
                return f"(N={self.visit_count}, Q_v={self.mean_value:.2f}, best={self.max_value:.2f}) [root]"

        import gymcts.colorful_console_utils as ccu

        if self.is_root():
            return f"({ccu.CYELLOW}N{ccu.CEND}={self.visit_count}, {ccu.CYELLOW}Q_v{ccu.CEND}={self.mean_value:.2f}, {ccu.CYELLOW}best{ccu.CEND}={self.max_value:.2f})"

        if action_space_n is None:
            raise ValueError("action_space_n must be provided if colored is True")

        p = ccu.CYELLOW
        e = ccu.CEND
        v = ccu.CCYAN

        def colorful_value(value: float | int | None) -> str:
            if value == None:
                return f"{ccu.CGREY}None{e}"
            color = ccu.CCYAN
            if value == 0:
                color = ccu.CRED
            if value == float("inf"):
                color = ccu.CGREY
            if value == -float("inf"):
                color = ccu.CGREY

            if isinstance(value, float):
                return f"{color}{value:.2f}{e}"

            if isinstance(value, int):
                return f"{color}{value}{e}"

        root_node = self.get_root()
        mean_val = f"{self.mean_value:.2f}"

        return ((f"("
                 f"{p}a{e}={ccu.wrap_evenly_spaced_color(s=self.action, n_of_item=self.action, n_classes=action_space_n)}, "
                 f"{p}N{e}={colorful_value(self.visit_count)}, "
                 f"{p}Q_v{e}={ccu.wrap_with_color_scale(s=mean_val, value=self.mean_value, min_val=root_node.min_value, max_val=root_node.max_value)}, "
                 f"{p}best{e}={colorful_value(self.max_value)}") +
                (f", {p}{GymctsNode.score_variate}{e}={colorful_value(self.tree_policy_score())})" if not self.is_root() else ")"))

    def traverse_nodes(self) -> Generator[TGymctsNode, None, None]:
        """
        Traverse the tree and yield all nodes in the tree.

        :return: a generator that yields all nodes in the tree.
        """
        yield self
        if self.children:
            for child in self.children.values():
                yield from child.traverse_nodes()

    def get_root(self) -> TGymctsNode:
        """
        Returns the root node of the tree. The root node is the node that has no parent.

        :return: the root node of the tree.
        """
        if self.is_root():
            return self
        return self.parent.get_root()

    def max_tree_depth(self):
        """
        Returns the maximum depth of the tree. The depth of a node is the number of edges from
        the node to the root node.

        :return: the maximum depth of the tree.
        """
        if self.is_leaf():
            return 0
        return 1 + max(child.max_tree_depth() for child in self.children.values())

    def n_children_recursively(self):
        """
        Returns the number of children of the node recursively. The number of children of a node is the number of
        children of the node plus the number of children of all children of the node.

        :return: the number of children of the node recursively.
        """
        if self.is_leaf():
            return 0
        return len(self.children) + sum(child.n_children_recursively() for child in self.children.values())

    def __init__(self,
                 action: int | None,
                 parent: TGymctsNode | None,
                 env_reference: GymctsABC,
                 ):
        """
        Initializes the node. The node is initialized with the state of the environment and the action that was taken to
        reach the node. The node is also initialized with the parent node and the environment reference.

        :param action: the action that was taken to reach the node. If the node is a root node, this parameter is None.
        :param parent: the parent node of the node. If the node is a root node, this parameter is None.
        :param env_reference: a reference to the environment. The environment is used to get the state of the node and the valid actions.
        """

        # field depending on whether the node is a root node or not
        self.action: int | None

        self.env_reference: GymctsABC
        self.parent: GymctsNode | None
        self.uuid = uuid.uuid4()

        if parent is None:
            self.action = None
            self.parent = None
            if env_reference.is_terminal():
                raise ValueError("Root nodes shall not be terminal.")
        else:
            if action is None:
                raise ValueError("action must be provided if parent is not None")

            self.action = action
            self.parent = parent  # not None

        # fields that are always initialized the same way
        self.terminal: bool = env_reference.is_terminal()

        from copy import copy
        self.state = env_reference.get_state()
        # log.debug(f"saving state of node '{str(self)}' to memory location: {hex(id(self.state))}")
        self.visit_count: int = 0

        self.mean_value: float = 0
        self.max_value: float = -float("inf")
        self.min_value: float = +float("inf")

        # safe valid action instead of calling the environment
        # this reduces the compute but increases the memory usage
        self.valid_actions: list[int] = env_reference.get_valid_actions()
        self.children: dict[int, GymctsNode] | None = None  # may be expanded later

    def reset(self) -> None:
        self.parent = None
        self.visit_count: int = 0

        self.mean_value: float = 0
        self.max_value: float = -float("inf")
        self.min_value: float = +float("inf")
        self.children: dict[int, GymctsNode] | None = None  # may be expanded later

        # just setting the children of the parent node to None should be enough to trigger garbage collection
        # however, we also set the parent to None to make sure that the parent is not referenced anymore
        if self.parent:
            self.parent.reset()

    def remove_parent(self) -> None:
        self.parent = None

        if self.parent is not None:
            self.parent.remove_parent()

    def is_root(self) -> bool:
        """
        Returns true if the node is a root node. A root node is a node that has no parent.

        :return: true if the node is a root node, false otherwise.
        """
        return self.parent is None

    def is_leaf(self) -> bool:
        """
        Returns true if the node is a leaf node. A leaf node is a node that has no children. A leaf node is a node that has no children.

        :return: true if the node is a leaf node, false otherwise.
        """
        return self.children is None or len(self.children) == 0

    def get_random_child(self) -> TGymctsNode:
        """
        Returns a random child of the node. A random child is a child that is selected randomly from the list of children.
        :return:
        """
        if self.is_leaf():
            raise ValueError("cannot get random child of leaf node")  # todo: maybe return self instead?

        return list(self.children.values())[random.randint(0, len(self.children) - 1)]

    def get_best_action(self) -> int:
        """
        Returns the best action of the node. The best action is the action that has the highest score.
        The score is calculated using the get_score() method. The best action is the action that has the highest score.
        The best action is the action that has the highest score.

        :return: the best action of the node.
        """
        return max(self.children.values(), key=lambda child: child.get_score()).action

    def get_score(self) -> float:  # todo: make it an attribute?
        """
        Returns the score of the node. The score is calculated using the mean value and the maximum value of the node.
        The score is calculated using the formula: score = (1 - a) * mean_value + a * max_value
        where a is the best action weight.

        :return: the score of the node.
        """
        # return self.mean_value
        assert 0 <= GymctsNode.best_action_weight <= 1
        a = GymctsNode.best_action_weight
        return (1 - a) * self.mean_value + a * self.max_value

    def get_mean_value(self) -> float:
        return self.mean_value

    def get_max_value(self) -> float:
        """
        Returns the maximum value of the node. The maximum value is the maximum value of the node.

        :return: the maximum value of the node.
        """
        return self.max_value

    def tree_policy_score(self):
        """
        TODO: update docstring

        The score for an action that would transition between the parent and child.
        For vanilla MCTS, this is the UCB1 score.

        The UCB1 score is calculated using the formula:

        UCT (Upper Confidence Bound applied to Trees) exploration terms:

        UCT_v0:
            c * √( 2 * ln(N(s)) / N(s,a) )

        UCT_v1:
            c * √( ln(N(s)) / (1 + N(s,a)) )

        UCT_v2:
            c * ( √(N(s)) / (1 + N(s,a)) )

        Where:
            N(s)     = number of times state s has been visited
            N(s,a)   = number of times action a was taken from state s
            c        = exploration constant

        where:
        - mean_value is the mean value of the node
        - c is a constant that controls the exploration-exploitation trade-off (GymctsNode.ubc_c)
        - parent_visit_count is the number of times the parent node has been visited
        - visit_count is the number of times the node has been visited

        If the node has not been visited yet, the score is set to infinity.

        prior_score = child.prior * math.sqrt(parent.visit_count) / (child.visit_count + 1)

        if child.visit_count > 0:
            # The value of the child is from the perspective of the opposing player
            value_score = -child.value()
        else:
            value_score = 0

            return value_score + prior_score

        :return:
        """
        if self.is_root():
            raise ValueError("ucb_score can only be called on non-root nodes")
        c = GymctsNode.ubc_c # default is 0.707

        assert 0 <= GymctsNode.best_action_weight <= 1
        b = GymctsNode.best_action_weight
        exploitation_term = 0.0 if self.visit_count == 0 else (1 - b) * self.mean_value + b * self.max_value

        if GymctsNode.score_variate == "UCT_v0":
            if self.visit_count == 0:
                return float("inf")
            return exploitation_term + c * math.sqrt( 2 * math.log(self.parent.visit_count) / (self.visit_count))

        if GymctsNode.score_variate == "UCT_v1":
            return exploitation_term + c * math.sqrt( math.log(self.parent.visit_count) / (1 + self.visit_count))

        if GymctsNode.score_variate == "UCT_v2":
            return exploitation_term + c * math.sqrt(self.parent.visit_count) / (1 + self.visit_count)

        raise ValueError(f"unknown score variate: {GymctsNode.score_variate}. ")






import copy
import sys
from typing import Any, Literal

import random
import math
import sb3_contrib

import gymnasium as gym
import numpy as np

from graph_jsp_env.disjunctive_graph_jsp_env import DisjunctiveGraphJspEnv
from jsp_instance_utils.instances import ft06, ft06_makespan
from sb3_contrib.common.maskable.distributions import MaskableCategoricalDistribution
from sb3_contrib.common.maskable.policies import MaskableActorCriticPolicy
from sb3_contrib.common.wrappers import ActionMasker

from gymcts.gymcts_agent import GymctsAgent
from gymcts.gymcts_env_abc import GymctsABC
from gymcts.gymcts_node import GymctsNode

from gymcts.logger import log


class GraphJspNeuralGYMCTSWrapper(GymctsABC, gym.Wrapper):

    def __init__(self, env: DisjunctiveGraphJspEnv):
        gym.Wrapper.__init__(self, env)

    def load_state(self, state: Any) -> None:
        self.env.reset()
        for action in state:
            self.env.step(action)

    def is_terminal(self) -> bool:
        return self.env.unwrapped.is_terminal()

    def get_valid_actions(self) -> list[int]:
        return list(self.env.unwrapped.valid_actions())

    def rollout(self) -> float:
        terminal = env.is_terminal()

        if terminal:
            lower_bound = env.unwrapped.reward_function_parameters['scaling_divisor']
            return - env.unwrapped.get_makespan() / lower_bound + 2

        reward = 0
        while not terminal:
            action = random.choice(self.get_valid_actions())
            obs, reward, terminal, truncated, _ = env.step(action)

        return reward + 2

    def get_state(self) -> Any:
        return env.unwrapped.get_action_history()

    def action_masks(self) -> np.ndarray | None:
        """Return the action mask for the current state."""
        return self.env.unwrapped.valid_action_mask()


class GymctsNeuralNode(GymctsNode):
    PUCT_v3_mu = 0.95

    MuZero_c1 = 1.25
    MuZero_c2 = 19652.0

    """
    PUCT (Predictor + UCT) exploration terms:

    PUCT_v0:
        c * P(s, a) * √( N(s) / (1 + N(s,a)) )

    PUCT_v1:
        c * P(s, a) * √( 2 * ln(N(s)) / N(s,a) )

    PUCT_v2:
        c * P(s, a) * √( N(s) ) / N(s,a)

    PUCT_v3:
        c * P(s, a)^μ * √( N(s) / (1 + N(s,a)) )

    PUCT_v4:
        c * ( P(s, a) / (1 + N(s,a)) )

    PUCT_v5:
        c * P(s, a) * ( √(N(s)) + 1 ) / (N(s,a) + 1)

    PUCT_v6:
        c * P(s, a) * N(s) / (1 + N(s,a))

    PUCT_v7:
        c * P(s, a) * ( √(N(s)) + ε ) / (N(s,a) + 1)

    PUCT_v8:
        c * P(s, a) * √( (ln(N(s)) + 1) / (1 + N(s,a)) )

    PUCT_v9:
        c * P(s, a) * √( N(s) / (1 + N(s,a)) )

    PUCT_v10:
        c * P(s, a) * √( ln(N(s)) / (1 + N(s,a)) )


    MuZero exploration terms:

    MuZero_v0:
        P(s, a) * √( N(s) / (1 + N(s,a)) ) * [ c₁ + ln( (N(s) + c₂ + 1) / c₂ ) ]

    MuZero_v1:
        P(s, a) * √( N(s) / (1 + N(s,a)) ) * [ c₁ + ln( (N(s) + c₂ + 1) / c₂ ) ]


    Where:
        - N(s):      number of times state s has been visited
        - N(s,a):    number of times action a was taken from state s
        - P(s,a):    prior probability of selecting action a from state s
        - c, c₁, c₂: exploration constants
        - μ:         exponent applied to P(s,a) in some variants
        - ε:         small constant to avoid division by zero (in PUCT 7)
    """
    score_variate: Literal[
        "PUCT_v0",
        "PUCT_v1",
        "PUCT_v2",
        "PUCT_v3",
        "PUCT_v4",
        "PUCT_v5",
        "PUCT_v6",
        "PUCT_v7",
        "PUCT_v8",
        "PUCT_v9",
        "PUCT_v10",
        "MuZero_v0",
        "MuZero_v1",
    ] = "PUCT_v0"

    def __init__(
            self,
            action: int,
            parent: 'GymctsNeuralNode',
            env_reference: GymctsABC,
            prior_selection_score: float,
            observation: np.ndarray | None = None,
    ):
        super().__init__(action, parent, env_reference)

        self._obs = observation
        self._selection_score_prior = prior_selection_score

    def tree_policy_score(self) -> float:
        # call the superclass (GymctsNode) for ucb_score
        c = GymctsNode.ubc_c
        # the way alpha zero does it
        # exploration_term = self._selection_score_prior * c * math.sqrt(math.log(self.parent.visit_count)) / (1 + self.visit_count)
        # the way the vanilla gymcts does it
        p_sa = self._selection_score_prior
        n_s = self.parent.visit_count
        n_sa = self.visit_count

        assert 0 <= GymctsNode.best_action_weight <= 1
        b = GymctsNode.best_action_weight
        exploitation_term = 0.0 if self.visit_count == 0 else (1 - b) * self.mean_value + b * self.max_value

        if GymctsNeuralNode.score_variate == "PUCT_v0":
            return exploitation_term + c * p_sa * math.sqrt(n_s) / (1 + n_sa)
        elif GymctsNeuralNode.score_variate == "PUCT_v1":
            return exploitation_term + c * p_sa * math.sqrt(2 * math.log(n_s) / (n_sa))
        elif GymctsNeuralNode.score_variate == "PUCT_v2":
            if n_sa == 0:
                return float("inf")  # Avoid division by zero
            return exploitation_term + c * p_sa * math.sqrt(n_s) / n_sa
        elif GymctsNeuralNode.score_variate == "PUCT_v3":
            return exploitation_term + c * (p_sa ** GymctsNeuralNode.PUCT_v3_mu) * math.sqrt(n_s / (1 + n_sa))
        elif GymctsNeuralNode.score_variate == "PUCT_v4":
            return exploitation_term + c * (p_sa / (1 + n_sa))
        elif GymctsNeuralNode.score_variate == "PUCT_v5":
            return exploitation_term + c * p_sa * (math.sqrt(n_s) + 1) / (n_sa + 1)
        elif GymctsNeuralNode.score_variate == "PUCT_v6":
            return exploitation_term + c * p_sa * n_s / (1 + n_sa)
        elif GymctsNeuralNode.score_variate == "PUCT_v7":
            epsilon = 1e-8
            return exploitation_term + c * p_sa * (math.sqrt(n_s) + epsilon) / (n_sa + 1)
        elif GymctsNeuralNode.score_variate == "PUCT_v8":
            return exploitation_term + c * p_sa * math.sqrt((math.log(n_s) + 1) / (1 + n_sa))
        elif GymctsNeuralNode.score_variate == "PUCT_v9":
            return exploitation_term + c * p_sa * math.sqrt(n_s / (1 + n_sa))
        elif GymctsNeuralNode.score_variate == "PUCT_v10":
            return exploitation_term + c * p_sa * math.sqrt(math.log(n_s) / (1 + n_sa))
        elif GymctsNeuralNode.score_variate == "MuZero_v0":
            c1 = GymctsNeuralNode.MuZero_c1
            c2 = GymctsNeuralNode.MuZero_c2
            return exploitation_term + c * p_sa * math.sqrt(n_s) / (1 + n_sa) * (c1 + math.log((n_s + c2 + 1) / c2))
        elif GymctsNeuralNode.score_variate == "MuZero_v1":
            c1 = GymctsNeuralNode.MuZero_c1
            c2 = GymctsNeuralNode.MuZero_c2
            return exploitation_term + c * p_sa * math.sqrt(n_s) / (1 + n_sa) * (c1 + math.log((n_s + c2 + 1) / c2))

        exploration_term = self._selection_score_prior * c * math.sqrt(
            math.log(self.parent.visit_count) / (self.visit_count)) if self.visit_count > 0 else float("inf")
        return self.mean_value + exploration_term

    def get_best_action(self) -> int:
        """
        Returns the best action of the node. The best action is the action with the highest score.
        The best action is the action that has the highest score.

        :return: the best action of the node.
        """
        return max(self.children.values(), key=lambda child: child.max_value).action

    def __str__(self, colored=False, action_space_n=None) -> str:
        """
        Returns a string representation of the node. The string representation is used for visualisation purposes.
        It is used for example in the mcts tree visualisation functionality.

        :param colored: true if the string representation should be colored, false otherwise. (ture is used by the mcts tree visualisation)
        :param action_space_n: the number of actions in the action space. This is used for coloring the action in the string representation.
        :return: a potentially colored string representation of the node.
        """
        if not colored:

            if not self.is_root():
                return f"(a={self.action}, N={self.visit_count}, Q_v={self.mean_value:.2f}, best={self.max_value:.2f}, {GymctsNeuralNode.score_variate}={self.tree_policy_score():.2f})"
            else:
                return f"(N={self.visit_count}, Q_v={self.mean_value:.2f}, best={self.max_value:.2f}) [root]"

        import gymcts.colorful_console_utils as ccu

        if self.is_root():
            return f"({ccu.CYELLOW}N{ccu.CEND}={self.visit_count}, {ccu.CYELLOW}Q_v{ccu.CEND}={self.mean_value:.2f}, {ccu.CYELLOW}best{ccu.CEND}={self.max_value:.2f})"

        if action_space_n is None:
            raise ValueError("action_space_n must be provided if colored is True")

        p = ccu.CYELLOW
        e = ccu.CEND
        v = ccu.CCYAN

        def colorful_value(value: float | int | None) -> str:
            if value == None:
                return f"{ccu.CGREY}None{e}"
            color = ccu.CCYAN
            if value == 0:
                color = ccu.CRED
            if value == float("inf"):
                color = ccu.CGREY
            if value == -float("inf"):
                color = ccu.CGREY

            if isinstance(value, float):
                return f"{color}{value:.2f}{e}"

            if isinstance(value, int):
                return f"{color}{value}{e}"

        root_node = self.get_root()
        mean_val = f"{self.mean_value:.2f}"

        return ((f"("
                 f"{p}a{e}={ccu.wrap_evenly_spaced_color(s=self.action, n_of_item=self.action, n_classes=action_space_n)}, "
                 f"{p}N{e}={colorful_value(self.visit_count)}, "
                 f"{p}Q_v{e}={ccu.wrap_with_color_scale(s=mean_val, value=self.mean_value, min_val=root_node.min_value, max_val=root_node.max_value)}, "
                 f"{p}best{e}={colorful_value(self.max_value)}") +
                (
                    f", {p}{GymctsNeuralNode.score_variate}{e}={colorful_value(self.tree_policy_score())})" if not self.is_root() else ")"))


class GymctsNeuralAgent(GymctsAgent):

    def __init__(self,
                 env: GymctsABC,
                 *args,
                 model_kwargs=None,
                 score_variate: Literal[
                     "PUCT_v0",
                     "PUCT_v1",
                     "PUCT_v2",
                     "PUCT_v3",
                     "PUCT_v4",
                     "PUCT_v5",
                     "PUCT_v6",
                     "PUCT_v7",
                     "PUCT_v8",
                     "PUCT_v9",
                     "PUCT_v10",
                     "MuZero_v0",
                     "MuZero_v1",
                 ] = "PUCT_v0",
                 **kwargs
                 ):

        # init super class
        super().__init__(
            env=env,
            *args,
            **kwargs
        )
        if score_variate not in [
            "PUCT_v0",
            "PUCT_v1",
            "PUCT_v2",
            "PUCT_v3",
            "PUCT_v4",
            "PUCT_v5",
            "PUCT_v6",
            "PUCT_v7",
            "PUCT_v8",
            "PUCT_v9",
            "PUCT_v10",
            "MuZero_v0",
            "MuZero_v1",
        ]:
            raise ValueError(f"Invalid score_variate: {score_variate}. Must be one of: "
                             f"['PUCT_v0', 'PUCT_v1', 'PUCT_v2', 'PUCT_v3', 'PUCT_v4', 'PUCT_v5', "
                             f"'PUCT_v6', 'PUCT_v7', 'PUCT_v8', 'PUCT_v9', 'PUCT_v10', 'MuZero_v0', 'MuZero_v1']")
        GymctsNeuralNode.score_variate = score_variate

        if model_kwargs is None:
            model_kwargs = {}
        obs, info = env.reset()

        self.search_root_node = GymctsNeuralNode(
            action=None,
            parent=None,
            env_reference=env,
            observation=obs,
            prior_selection_score=1.0,
        )

        def mask_fn(env: gym.Env) -> np.ndarray:
            mask = env.action_masks()
            if mask is None:
                mask = np.ones(env.action_space.n, dtype=np.float32)
            return mask

        env = ActionMasker(env, action_mask_fn=mask_fn)

        model_kwargs = {
                           "policy": MaskableActorCriticPolicy,
                           "env": env,
                           "verbose": 1,
                       } | model_kwargs

        self._model = sb3_contrib.MaskablePPO(**model_kwargs)

    def learn(self, total_timesteps: int, **kwargs) -> None:
        """Learn from the environment using the MaskablePPO model."""
        self._model.learn(total_timesteps=total_timesteps, **kwargs)

    def expand_node(self, node: GymctsNeuralNode) -> None:
        log.debug(f"expanding node: {node}")
        # EXPANSION STRATEGY
        # expand all children

        child_dict = {}

        self._load_state(node)

        obs_tensor, vectorized_env = self._model.policy.obs_to_tensor(np.array([node._obs]))
        action_masks = np.array([self.env.action_masks()])
        distribution = self._model.policy.get_distribution(obs=obs_tensor, action_masks=action_masks)
        unwrapped_distribution = distribution.distribution.probs[0]

        # print(f'valid actions: {node.valid_actions}')
        # print(f'env mask: {self.env.action_masks()}')
        # print(f'env valid actions: {self.env.get_valid_actions()}')
        """
                for action in node.valid_actions:
            # reconstruct state
            # load state of leaf node
            self._load_state(node)

            obs, reward, terminal, truncated, _ = self.env.step(action)
            child_dict[action] = GymctsNeuralNode(
                action=action,
                parent=node,
                env_reference=self.env,
                observation=obs,
                prior_selection_score=1.0,
            )
        node.children = child_dict
        return
        """

        for action, prob in enumerate(unwrapped_distribution):
            self._load_state(node)

            log.debug(f"Probabily for action {action}: {prob}")

            if prob == 0.0:
                continue

            assert action in node.valid_actions, f"Action {action} is not in valid actions: {node.valid_actions}"

            obs, reward, terminal, truncated, _ = self.env.step(action)
            child_dict[action] = GymctsNeuralNode(
                action=action,
                parent=node,
                observation=copy.deepcopy(obs),
                env_reference=self.env,
                prior_selection_score=float(prob)
            )

        node.children = child_dict
        # print(f"Expanded node {node} with {len(node.children)} children.")


if __name__ == '__main__':
    log.setLevel(20)

    env_kwargs = {
        "jps_instance": ft06,
        "default_visualisations": ["gantt_console", "graph_console"],
        "reward_function_parameters": {
            "scaling_divisor": ft06_makespan
        },
        "reward_function": "nasuta",
    }

    env = DisjunctiveGraphJspEnv(**env_kwargs)
    env.reset()

    env = GraphJspNeuralGYMCTSWrapper(env)

    import torch

    model_kwargs = {
        "gamma": 0.99013,
        "gae_lambda": 0.9,
        "normalize_advantage": True,
        "n_epochs": 28,
        "n_steps": 432,
        "max_grad_norm": 0.5,
        "learning_rate": 6e-4,
        "policy_kwargs": {
            "net_arch": {
                "pi": [90, 90],
                "vf": [90, 90],
            },
            "ortho_init": True,
            "activation_fn": torch.nn.ELU,
            "optimizer_kwargs": {
                "eps": 1e-7
            }
        }
    }

    agent = GymctsNeuralAgent(
        env=env,
        render_tree_after_step=True,
        render_tree_max_depth=3,
        exclude_unvisited_nodes_from_render=False,
        number_of_simulations_per_step=15,
        # clear_mcts_tree_after_step = False,
        model_kwargs=model_kwargs
    )

    agent.learn(total_timesteps=10_000)

    agent.solve()

    actions = agent.solve(render_tree_after_step=True)
    for a in actions:
        obs, rew, term, trun, info = env.step(a)

    env.render()
    makespan = env.unwrapped.get_makespan()
    print(f"makespan: {makespan}")



from gymcts.gymcts_agent import GymctsAgent


def test_number_of_visits_without_clearing_root(graph_matrix_env_naive_wrapper_singe_job_jsp_instance):
    env = graph_matrix_env_naive_wrapper_singe_job_jsp_instance

    agent = GymctsAgent(env=env, clear_mcts_tree_after_step=False, keep_whole_tree_till_initial_root=True)

    assert agent.search_root_node.visit_count == 0
    agent.vanilla_mcts_search(search_start_node=agent.search_root_node, num_simulations=10)

    assert agent.search_root_node.visit_count == 10

    agent.vanilla_mcts_search(search_start_node=agent.search_root_node, num_simulations=15)

    assert agent.search_root_node.visit_count == 25



def test_number_of_visits_without_clearing(graph_matrix_env_naive_wrapper_singe_job_jsp_instance):
    env = graph_matrix_env_naive_wrapper_singe_job_jsp_instance

    agent = GymctsAgent(env=env, clear_mcts_tree_after_step=False, keep_whole_tree_till_initial_root=True)
    assert agent.search_root_node.visit_count == 0

    actions = agent.solve(num_simulations_per_step=10)

    assert len(actions) == 4

    assert agent.search_root_node.get_root().visit_count == 10 * 4

    # check visit counts for all nodes
    for node, counts, tree_depth in zip(
            agent.search_root_node.get_root().traverse_nodes(),
            [40, 39, 38, 37],
            [4, 3, 2, 1]
    ):
        assert node.visit_count == counts
        assert node.max_tree_depth() == tree_depth


def test_number_of_visits_without_clearing_root_dynamic_step_size(graph_matrix_env_naive_wrapper_singe_job_jsp_instance):

    env = graph_matrix_env_naive_wrapper_singe_job_jsp_instance

    agent = GymctsAgent(env=env, clear_mcts_tree_after_step=False, keep_whole_tree_till_initial_root=True)

    tree_root = agent.search_root_node

    assert agent.search_root_node.visit_count == 0
    agent.perform_mcts_step(search_start_node=agent.search_root_node, num_simulations=10)

    assert agent.search_root_node.visit_count == 10 - 1 # -1 because the root node is not visited one before expanding
    assert agent.search_root_node.get_root() == tree_root

    agent.perform_mcts_step(search_start_node=agent.search_root_node, num_simulations=15)

    assert agent.search_root_node.visit_count == (10 - 1) + 15 - 1
    assert agent.search_root_node.get_root() == tree_root


def test_number_of_visits_with_clearing_root_dynamic_step_size(graph_matrix_env_naive_wrapper_singe_job_jsp_instance):

    env = graph_matrix_env_naive_wrapper_singe_job_jsp_instance

    agent = GymctsAgent(env=env, clear_mcts_tree_after_step=True)

    assert agent.search_root_node.visit_count == 0
    action, node = agent.perform_mcts_step(search_start_node=agent.search_root_node, num_simulations=10)
    assert agent.search_root_node == node
    assert agent.search_root_node.is_root()

    assert agent.search_root_node.visit_count == 0

    action, node = agent.perform_mcts_step(search_start_node=agent.search_root_node, num_simulations=15)
    assert agent.search_root_node.visit_count == 0
    assert agent.search_root_node == node
    assert agent.search_root_node.is_root()




def test_number_of_visits_with_clearing_root2(graph_matrix_env_naive_wrapper_two_job_jsp_instance):
    env = graph_matrix_env_naive_wrapper_two_job_jsp_instance

    steps = 50 # total of step to get an exhaustive tree is 2**4+1 = 16 +1 = 17
    # this means after the first step the tree should not grow anymore

    agent = GymctsAgent(
        env=env,
        clear_mcts_tree_after_step=False,
        keep_whole_tree_till_initial_root=True,
        number_of_simulations_per_step=50
    )


    for i in range(8):
        action, node = agent.perform_mcts_step()
        assert node.get_root().max_tree_depth() == 8
        assert node.get_root().visit_count == (i+1) * 50

    assert node.terminal











import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.INFO,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(show_path=False)]
)

log = logging.getLogger("rich")

banner_sw = f"""

      ▟█████▛▜█▙▟█▛ ▟█▙  ▟██▛▟█████▛████▛▟████▛
     ▟█▛      ▜██▛ ▟█▛██▛██▛▟█▛     ▟█▛  ▜███▙ 
    ▟█▛ ▟█▛   ▟█▛ ▟█▛   ▟█▛▟█▛     ▟█▛      ▟█▛ 
    ▜████▛   ▟█▛ ▟█▛   ▟█▛ ▜████▛ ▟█▛  ▟████▛  
                                           
"""


if __name__ == '__main__':
    log.debug("Hello, World!")
    log.info("Hello, World!")
    log.error("Hello, World!")
    log.warning("Hello, World!")
    log.critical("Hello, World!")



from typing import Any

import matplotlib.pyplot as plt
import numpy as np

CEND = "\33[0m"
CBOLD = "\33[1m"
CITALIC = "\33[3m"
CURL = "\33[4m"
CBLINK = "\33[5m"
CBLINK2 = "\33[6m"
CSELECTED = "\33[7m"

CBLACK = "\33[30m"
CRED = "\33[31m"
CGREEN = "\33[32m"
CYELLOW = "\33[33m"
CBLUE = "\33[34m"
CCYAN = '\33[96m'
CMAGENTA = '\033[35m'
CVIOLET = "\33[35m"
CBEIGE = "\33[36m"
CWHITE = "\33[37m"

CBLACKBG = "\33[40m"
CREDBG = "\33[41m"
CGREENBG = "\33[42m"
CYELLOWBG = "\33[43m"
CBLUEBG = "\33[44m"
CVIOLETBG = "\33[45m"
CBEIGEBG = "\33[46m"
CWHITEBG = "\33[47m"

CGREY = "\33[90m"
CRED2 = "\33[91m"
CGREEN2 = "\33[92m"
CYELLOW2 = "\33[93m"
CBLUE2 = "\33[94m"
CCYAN2 = "\033[36m"
CVIOLET2 = "\33[95m"
CBEIGE2 = "\33[96m"
CWHITE2 = "\33[97m"

CGREYBG = "\33[100m"
CREDBG2 = "\33[101m"
CGREENBG2 = "\33[102m"
CYELLOWBG2 = "\33[103m"
CBLUEBG2 = "\33[104m"
CVIOLETBG2 = "\33[105m"
CBEIGEBG2 = "\33[106m"
CWHITEBG2 = "\33[107m"


def rgb_color_sequence(r: int | float, g: int | float, b: int | float,
                       *, format_type: str = 'foreground') -> str:
    """
    generates a color-codes, that change the color of text in console outputs.

    rgb values must be numbers between 0 and 255 or 0.0 and 1.0.

    :param r:               red value.
    :param g:               green value
    :param b:               blue value

    :param format_type:     specifies weather the foreground-color or the background-color shall be adjusted.
                            valid options: 'foreground','background'
    :return:                a string that contains the color-codes.
    """
    # type: ignore # noqa: F401
    if format_type == 'foreground':
        f = '\033[38;2;{};{};{}m'.format  # font rgb format
    elif format_type == 'background':
        f = '\033[48;2;{};{};{}m'.format  # font background rgb format
    else:
        raise ValueError(f"format {format_type} is not defined. Use 'foreground' or 'background'.")
    rgb = [r, g, b]

    if isinstance(r, int) and isinstance(g, int) and isinstance(b, int):
        if min(rgb) < 0 and max(rgb) > 255:
            raise ValueError("rgb values must be numbers between 0 and 255 or 0.0 and 1.0")
        return f(r, g, b)
    if isinstance(r, float) and isinstance(g, float) and isinstance(b, float):
        if min(rgb) < 0 and max(rgb) > 1:
            raise ValueError("rgb values must be numbers between 0 and 255 or 0.0 and 1.0")
        return f(*[int(n * 255) for n in [r, g, b]])


def wrap_with_color_codes(s: object, /, r: int | float, g: int | float, b: int | float, **kwargs) \
        -> str:
    """
    stringify an object and wrap it with console color codes. It adds the color control sequence in front and one
    at the end that resolves the color again.

    rgb values must be numbers between 0 and 255 or 0.0 and 1.0.

    :param s: the object to stringify and wrap
    :param r: red value.
    :param g: green value.
    :param b: blue value.
    :param kwargs: additional argument for the 'DisjunctiveGraphJspVisualizer.rgb_color_sequence'-method.
    :return:
    """
    return f"{rgb_color_sequence(r, g, b, **kwargs)}" \
           f"{s}" \
           f"{CEND}"


def wrap_evenly_spaced_color(s: Any, n_of_item: int, n_classes: int, c_map="rainbow") -> str:
    """
    Wraps a string with a color scale (a matplotlib c_map) based on the n_of_item and n_classes.
    This function is used to color code the available actions in the MCTS tree visualisation.
    The children of the MCTS tree are colored based on their action for a clearer visualisation.

    :param s: the string (or object) to be wrapped. objects are converted to string (using the __str__ function).
    :param n_of_item: the index of the item to be colored. In a mcts tree, this is the (parent-)action of the node.
    :param n_classes: the number of classes (or items) to be colored. In a mcts tree, this is the number of available actions.
    :param c_map: the colormap to be used (default is 'rainbow').
                  The colormap can be any matplotlib colormap, e.g. 'viridis', 'plasma', 'inferno', 'magma', 'cividis'.
    :return: a string that contains the color-codes (prefix and suffix) and the string s in between.
    """
    if s is None or n_of_item is None or n_classes is None:
        return s

    c_map = plt.cm.get_cmap(c_map)  # select the desired cmap
    arr = np.linspace(0, 1, n_classes + 1)  # create a list with numbers from 0 to 1 with n items

    color_vals = c_map(arr[n_of_item])[:-1]
    color_asni = rgb_color_sequence(*color_vals, format_type='foreground')

    return f"{color_asni}{s}{CEND}"


def wrap_with_color_scale(s: str, value: float, min_val: float, max_val: float, c_map=None) -> str:
    """
    Wraps a string with a color scale (a matplotlib c_map) based on the value, min_val, and max_val.

    :param s: the string to be wrapped
    :param value: the value to be mapped to a color
    :param min_val: the minimum value of the scale
    :param max_val: the maximum value of the scale
    :param c_map: the colormap to be used (default is 'rainbow')
    :return:
    """
    if s is None or min_val is None or max_val is None or min_val >= max_val:
        return s

    if c_map is not None:
        c_map = plt.cm.get_cmap(c_map)  # select the desired cmap
    else:
        from matplotlib.colors import LinearSegmentedColormap
        colors = [
            np.array([255 / 255, 100 / 255, 128 / 255, 1.0]),  # RGBA values
            np.array([63 / 255, 197 / 255, 161 / 255, 1.0]),  # RGBA values
        ]
        c_map = LinearSegmentedColormap.from_list("custom_cmap", colors, N=256)

    color_vals = c_map((value - min_val) / (max_val - min_val))[:-1]
    color_asni = rgb_color_sequence(*color_vals, format_type='foreground')

    return f"{color_asni}{s}{CEND}"


if __name__ == '__main__':
    res = wrap_with_color_scale("test", 1.0, 0, 1)
    print(res)
